@@ -13,16 +13,19 @@
 // License for the specific language governing permissions and limitations under
 // the License.
 
-use std::{ffi, io, net, path};
+use std::{ffi, io, path, time};
 
 use clap::Parser;
 use is_terminal::IsTerminal;
 
-use crate::error::Error;
+use crate::{
+    conn_expr::{ConnectionExpr, ConnectionExprSource},
+    error::Error,
+};
 
 #[derive(Debug)]
 pub struct Args {
-    pub host: HostArg,
+    pub conn_expr_src: ConnectionExprSource,
     pub stdin_from: Option<IoArg>,
     pub stdout_to: Option<IoArg>,
     pub stderr_to: Option<IoArg>,
@@ -98,6 +101,8 @@ impl TryFrom<Cli> for Args {
             vec![IoArg::parse_from_path_or_pipe(f)
                 .map(SourceArg::from)
                 .map_err(|_| Error::BadSourceFile)?]
+        } else if cli.wait_port_file.is_some() {
+            vec![]
         } else {
             return Err(Error::NoInput);
         };
@@ -130,14 +135,17 @@ impl TryFrom<Cli> for Args {
             )
             .collect::<Result<_, _>>()?;
 
-        let host = if let Some(ref h) = cli.port {
-            HostArg::HostExpr(h.clone())
+        let conn_expr_src = if let Some(ref h) = cli.port {
+            h.into()
         } else {
-            HostArg::PortFile(cli.port_file.clone())
+            ConnectionExprSource::PortFile {
+                path: cli.port_file.clone(),
+                wait_for: cli.wait_port_file.map(time::Duration::from_secs),
+            }
         };
 
         Ok(Self {
-            host,
+            conn_expr_src,
             stdin_from,
             stdout_to: if cli.no_stdout {
                 None
@@ -249,8 +257,13 @@ pub struct IoParseError(String);
 )]
 struct Cli {
     /// Connect to server on [HOST:]PORT
-    #[clap(long, short, visible_alias = "host", value_name = "[HOST:]PORT")]
-    port: Option<HostExpr>,
+    #[clap(
+        long,
+        short,
+        visible_alias = "host",
+        value_name = "[[[USER@]TUNNEL[:PORT]:]HOST:]PORT"
+    )]
+    port: Option<ConnectionExpr>,
 
     /// Read server port from FILE
     #[clap(long, value_name = "FILE")]
@@ -332,13 +345,9 @@ struct Cli {
     )]
     _shebang_guard: bool,
 
-    /// Wait for .nrepl-port file
-    #[clap(
-        long = "wait-port-file",
-        value_name = "SECONDS",
-        parse(try_from_str = not_implemented),
-    )]
-    _wait_port_file: Option<u32>,
+    /// Wait .nrepl-port file to appear for SECONDS
+    #[clap(long = "wait-port-file", value_name = "SECONDS")]
+    wait_port_file: Option<u64>,
 
     /// Set timeout for program execution
     #[clap(
@@ -352,68 +361,3 @@ struct Cli {
 fn not_implemented<T>(_: &str) -> Result<T, &'static str> {
     Err("this feature has not been implemented yet, sorry")
 }
-
-#[derive(Debug)]
-pub enum HostArg {
-    HostExpr(HostExpr),
-    PortFile(Option<path::PathBuf>),
-}
-
-#[derive(Clone, Debug, PartialEq)]
-pub enum HostExpr {
-    Port(u16),
-    DomainPort(String, u16),
-    SocketAddr(net::SocketAddr),
-}
-
-impl std::str::FromStr for HostExpr {
-    type Err = HostOptionParseError;
-
-    fn from_str(s: &str) -> Result<Self, Self::Err> {
-        if let Ok(port) = s.parse::<u16>() {
-            return Ok(HostExpr::Port(port));
-        }
-        if let Ok(socket_addr) = s.parse::<net::SocketAddr>() {
-            return Ok(HostExpr::SocketAddr(socket_addr));
-        }
-        if let Some((host_part, port_part)) = s.rsplit_once(':') {
-            if let Ok(port) = port_part.parse::<u16>() {
-                if let Ok(domain) = addr::parse_domain_name(host_part) {
-                    return Ok(HostExpr::DomainPort(domain.to_string(), port));
-                }
-            }
-        }
-        Err(HostOptionParseError)
-    }
-}
-
-#[derive(Debug, PartialEq, thiserror::Error)]
-#[error("bad host and port expression")]
-pub struct HostOptionParseError;
-
-#[cfg(test)]
-mod test_options {
-
-    use super::*;
-    use std::net;
-
-    #[test]
-    fn host_option_parsing() {
-        assert_eq!("5678".parse::<HostExpr>(), Ok(HostExpr::Port(5678)),);
-        assert_eq!(
-            "1.2.3.4:5678".parse::<HostExpr>(),
-            Ok(HostExpr::SocketAddr(net::SocketAddr::from((
-                [1, 2, 3, 4],
-                5678
-            )))),
-        );
-        assert_eq!(
-            "localhost:5678".parse::<HostExpr>(),
-            Ok(HostExpr::DomainPort("localhost".to_string(), 5678)),
-        );
-        assert_eq!(
-            "example.com:5678".parse::<HostExpr>(),
-            Ok(HostExpr::DomainPort("example.com".to_string(), 5678)),
-        );
-    }
-}
